// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>


#include "enter.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <wctype.h>

#define MIN(a, b) ((a) < (b) ? (a) : (b))


/* redraw flags for mutt_enter_string() */
enum {
	M_REDRAW_INIT = 1, /* recalculate lengths */
	M_REDRAW_LINE,     /* redraw entire line */
	M_REDRAW_EOL,      /* redraw from current position to eol */
	M_REDRAW_PREV_EOL  /* redraw from curpos-1 to eol */
};

/* Returns:
 *	1 need to redraw the screen and call me again
 *	0 if input was given
 * 	-1 if abort.
 */
char * mutt_enter_string(const char * init) {
	int y, x;
	getyx(stdscr, y, x);
	km_dokey_load_termios();

	size_t curpos     = 0;      /* the location of the cursor */
	size_t lastchar   = 0;      /* offset of the last char in the string */
	size_t begin      = 0;      /* first character displayed on the line */
	int width;                  /* width of field */
	int redraw = M_REDRAW_INIT; /* when/what to redraw */
	size_t j;
	unsigned char * ret = NULL;

	for(;;) {
		if(redraw) {
			width = COLS - x - 1;

			if(redraw == M_REDRAW_INIT) {
				/* full redraw */
				lastchar = curpos = strlen(init ?: "");
				ret               = (unsigned char *)strdup(init ?: "");
				if(!ret)
					return NULL;

				begin = (ssize_t)lastchar > width ? lastchar - width : 0;
			}
			switch(redraw) {
				case M_REDRAW_PREV_EOL:
					j = curpos - 1;
					break;
				case M_REDRAW_EOL:
					j = curpos;
					break;
				default:
					j = begin;
			}
			move(y, x + j - begin);
			for(; j < lastchar && j < begin + width; j++)
				addch(ret[j]);
			clrtoeol();
			if(redraw != M_REDRAW_INIT)
				move(y, x + curpos - begin);
			redraw = 0;
		}
		refresh();

		/* first look to see if a keypress is an editor operation.  km_dokey()
		 * returns 0 if there is no entry in the keymap, so restore the last
		 * keypress and continue normally.
		 */
		wint_t LastKey;
		switch(km_dokey(&LastKey)) {
			case -1:
				free(ret);
				return NULL;

			case OP_EDITOR_BACKSPACE:
				if(curpos == 0) {
					beep();
					break;
				}
				for(j = curpos; j < lastchar; ++j)
					ret[j - 1] = ret[j];
				--curpos;
				--lastchar;
				if(curpos > begin) {
					if(lastchar == curpos) {
						move(y, x + curpos - begin);
						delch();
					} else
						redraw = M_REDRAW_EOL;
				} else {
					begin -= MIN(begin, (size_t)width / 2);
					redraw = M_REDRAW_LINE;
				}
				break;

			case OP_EDITOR_BOL:
				/* reposition the cursor at the beginning of the line */
				curpos = 0;
				if(begin) {
					/* the first char is not displayed, so readjust */
					begin  = 0;
					redraw = M_REDRAW_LINE;
				} else
					move(y, x);
				break;

			case OP_EDITOR_EOL:
				curpos = lastchar;
				if(lastchar < begin + width)
					move(y, x + lastchar - begin);
				else {
					begin  = lastchar - (size_t)width / 2;
					redraw = M_REDRAW_LINE;
				}
				break;

			case OP_EDITOR_KILL_LINE:
				lastchar = curpos = 0;
				begin             = 0;
				redraw            = M_REDRAW_LINE;
				break;

			case OP_EDITOR_KILL_EOL:
				lastchar = curpos;
				clrtoeol();
				break;

			case OP_EDITOR_BACKWARD_CHAR:
				if(curpos == 0) {
					beep();
				} else {
					curpos--;
					if(curpos < begin) {
						begin -= MIN(begin, (size_t)width / 2);
						redraw = M_REDRAW_LINE;
					} else
						move(y, x + curpos - begin);
				}
				break;

			case OP_EDITOR_FORWARD_CHAR:
				if(curpos == lastchar) {
					beep();
				} else {
					curpos++;
					if(curpos >= begin + width) {
						begin  = curpos - (size_t)width / 2;
						redraw = M_REDRAW_LINE;
					} else
						move(y, x + curpos - begin);
				}
				break;

			case OP_EDITOR_DELETE_CHAR:
				if(curpos != lastchar) {
					for(j = curpos; j < lastchar; ++j)
						ret[j] = ret[j + 1];
					--lastchar;
					redraw = M_REDRAW_EOL;
				} else
					beep();
				break;

			case OP_EDITOR_KILL_WORD:
				/* delete to begining of word */
				if(curpos != 0) {
					j = curpos;
					while(j > 0 && isspace(ret[j - 1]))
						j--;
					if(j > 0) {
						if(isalnum(ret[j - 1])) {
							for(j--; j > 0 && isalnum(ret[j - 1]); j--)
								;
						} else
							j--;
					}
					size_t save = j; /* save current position */
					while(curpos < lastchar)
						ret[j++] = ret[curpos++];
					lastchar = j;
					curpos   = save; /* restore current position */
					                 /* update screen */
					if(curpos < begin) {
						begin  = curpos - MIN(curpos, (size_t)width / 2);
						redraw = M_REDRAW_LINE;
					} else
						redraw = M_REDRAW_EOL;
				}
				break;

			case OP_EDITOR_IDK:
				beep();
				break;

			case OP_EDITOR_DONE:
				ret[lastchar] = 0;
				return (char *)ret;

			case OP_EDITOR_RESIZE:
				redraw = M_REDRAW_LINE;
				break;

			case OP_EDITOR_NOP:
				break;

			case OP_EDITOR_CHAR:
				if(iswprint(LastKey)) {
					char new[MB_CUR_MAX];
					int newlen = wctomb(new, LastKey);
					if(newlen == -1)
						goto err;

					if(!(ret = realloc(ret, lastchar + newlen + 1)))
						return NULL;
					memmove(ret + curpos + newlen, ret + curpos, lastchar - curpos);
					memcpy(ret + curpos, new, newlen);
					curpos += newlen;
					lastchar += newlen;

					if(curpos >= begin + width) {
						begin  = curpos - (size_t)width / 2;
						redraw = M_REDRAW_LINE;
					} else if(curpos == lastchar)
						addnstr(new, newlen);
					else
						redraw = M_REDRAW_PREV_EOL;
				} else {
				err:
					flushinp();
					beep();
				}
				break;
		}
	}
}
