// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright (C) 1997 Michael R. Elkins <me@cs.hmc.edu>
// Created:       Thu Dec  4 02:13:11 PST 1997
// Last Modified: Tue Jul  4 11:23:49 CEST 2000


#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <locale.h>
#include <ncurses.h>
#include <regex.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>
#if __has_include(<libintl.h>)
#include <libintl.h>
#else
#define gettext(s) (s)
#define ngettext(s, p, n) (n == 1 ? s : p)
#endif

#include "enter.h"
#include "quote.h"
#define MAX(a, b) ((a) > (b) ? (a) : (b))


enum redraw { NO, FULL, INDEX, MOTION };
struct url {
	char * url;
	int cursor_y, cursor_x;
	int last_y;
};

static char * regex_error(const regex_t * rx, int err) {
	size_t es = regerror(err, rx, NULL, 0);
	char * eb = malloc(es);
	if(eb)
		regerror(err, rx, eb, es);
	return eb;
}

static void search_common(const char * search, regex_t * rx, size_t urlcount, struct url * url, enum redraw * redraw, size_t * current,
                          void (*matcher)(regex_t * rx, size_t urlcount, struct url * url, size_t * current)) {
	if(search) {
		regex_t newrx;
		int regerr = regcomp(&newrx, search, REG_EXTENDED | REG_ICASE | REG_NOSUB);
		if(regerr) {
			char * err = regex_error(&newrx, regerr);
			mvaddstr(LINES - 1, 0, search);
			addstr(": ");
			addstr(err ?: "?");
			clrtoeol();
			free(err);
			*redraw = MOTION;
			return;
		}

		if(rx->re_nsub != (size_t)-1)
			regfree(rx);
		*rx = newrx;
	}

	matcher(rx, urlcount, url, current);

	move(LINES - 1, 0);
	clrtoeol();
	*redraw = MOTION;
}

static void search_forward(regex_t * rx, size_t urlcount, struct url * url, size_t * current) {
	for(size_t i = *current + 1; i < urlcount; ++i)
		if(regexec(rx, url[i].url, 0, NULL, 0) == 0) {
			*current = i;
			return;
		}
}

static void search_backward(regex_t * rx, size_t urlcount, struct url * url, size_t * current) {
	(void)urlcount;

	for(size_t i = *current - 1; i != (size_t)-1; --i)
		if(regexec(rx, url[i].url, 0, NULL, 0) == 0) {
			*current = i;
			return;
		}
}

int main(int argc, char ** argv) {
	setlocale(LC_ALL, "");
#if __has_include(<libintl.h>)
	bindtextdomain("urlview-ng", TEXTDOMAIN_DIRNAME);
	textdomain("urlview-ng");
#endif

	/*** read the initialization file ***/
	char * command     = NULL;
	char * regexp      = NULL;
	bool expert        = false;
	bool skip_browser  = false;
	bool menu_wrapping = false;
	bool edit_url      = false;
	bool mouse         = true;

	FILE * fp   = NULL;
	char * home = getenv("HOME");
	if(home) {
		char * fpp;
		if(asprintf(&fpp, "%s/.urlview", home) != -1) {
			fp = fopen(fpp, "re");
			free(fpp);
		}
	}
	if(!fp)
		fp = fopen(SYSTEM_INITFILE, "re");

	char * line    = NULL;
	size_t linecap = 0;
	if(fp) {
		for(ssize_t len; (len = getline(&line, &linecap, fp)) != -1;) {
			if(line[0] == '#' || line[0] == '\n')
				continue;
			if(line[len - 1] == '\n')
				line[--len] = '\0';

			size_t keylen   = strcspn(line, " \f\n\r\t\v");
			size_t seplen   = strspn(line + keylen, " \f\n\r\t\v");
			char * key      = line;
			char * value    = line + keylen + seplen;
			size_t valuelen = len - (keylen + seplen);

#define BOOLKEY(name, var)                                                                   \
	if(KEYMATCH(name)) {                                                                       \
		if(!strcasecmp("YES", value))                                                            \
			var = true;                                                                            \
		else if(!strcasecmp("NO", value))                                                        \
			var = false;                                                                           \
		else /* "BROWSER", user's input */                                                       \
			errx(1, gettext("Unknown value for %s: %s. Valid values are: yes, no."), name, value); \
	}

#define KEYMATCH(k) keylen == sizeof(k) - 1 && !memcmp(key, k, sizeof(k) - 1)
			if(KEYMATCH("REGEXP") && valuelen) {
				if(!(regexp = realloc(regexp, valuelen)))
					err(1, NULL);
				char * wc = regexp;
				while(*value) {
					if(*value == '\\') {
						value++;
						switch(*value) {
							case 'n':
								*wc++ = '\n';
								break;
							case 'r':
								*wc++ = '\r';
								break;
							case 't':
								*wc++ = '\t';
								break;
							case 'f':
								*wc++ = '\f';
								break;
							default:
								*wc++ = '\\';
								*wc++ = *value;
								break;
						}
					} else
						*wc++ = *value;
					++value;
				}
				*wc = '\0';
			} else if(KEYMATCH("COMMAND") && valuelen) {
				free(command);
				if(!(command = strdup(value)))
					err(1, NULL);
				skip_browser = true;
			} else
				BOOLKEY("WRAP", menu_wrapping)
			else                              //
			    BOOLKEY("EDIT", edit_url)     //
			    else                          //
			    BOOLKEY("MOUSE", mouse)       //
			    else if(KEYMATCH("BROWSER"))  //
			    skip_browser = false;
			else if(KEYMATCH("EXPERT"))  //
			    expert = true;
			else  //
			    errx(1, gettext("Unknown setting: %s"), line);
		}
		fclose(fp);
	}

	/* Only use the $BROWSER environment variable if
	 * (a) no COMMAND in rc file or
	 * (b) BROWSER in rc file.
	 * If both COMMAND and BROWSER are in the rc file, then the option used
	 * last counts.
	 */
	if(!skip_browser) {
		char * browser = getenv("BROWSER");
		if(browser) {
			if(strlen(browser) > 0)
				command = browser;
			else
				warnx(gettext("Empty $BROWSER, falling back to COMMAND (%s)."), command ?: DEFAULT_COMMAND);
		}
	}
	if(!command)
		command = (char *)DEFAULT_COMMAND;

	if(!expert && strchr(command, '\''))
		errx(1, gettext("\n"
		                "Your COMMAND/$BROWSER contains a single quote (') character.\n"
		                "This is most likely in error; please read the manual page for details.\n"
		                "If you really want to use this command, please put the word EXPERT\n"
		                "into a line of its own in your ~/.urlview file."));

	/*** compile the regexp ***/
	regex_t rx;
	int regerr = regcomp(&rx, regexp ?: DEFAULT_REGEXP, REG_EXTENDED | REG_ICASE | REG_NEWLINE);
	if(regerr)
		errx(1, "%s: %s", regexp ?: DEFAULT_REGEXP, regex_error(&rx, regerr));
	free(regexp);

	/*** prepare arguments ***/
	int startline   = 0;
	bool reopen_tty = false;
	size_t current = 0, oldcurrent = 0;

	opterr = false;
	for(int opt; (opt = getopt(argc, argv, "+0::1::2::3::4::5::6::7::8::9::")) != -1;) {
		if(opt == '?') {
			--optind;
			break;
		}

		if(optarg)
			startline = atoi(optarg - 1);
		else
			startline = opt - '0';
		current = (size_t)-1;
	}
	if(optind == argc)
		argv[--optind] = "-";

	/*** find matching patterns ***/
	struct url * url = NULL;
	size_t urlsize   = 0;
	size_t urlcount  = 0;
	bool error       = false;
	for(int i = optind; i < argc; ++i) {
		if(!strcmp("-", argv[i])) {
			fp         = stdin;
			reopen_tty = true;
		} else if(!(fp = fopen(argv[i], "re"))) {
			warn("%s", argv[i]);
			error = true;
			continue;
		}

		while(getline(&line, &linecap, fp) != -1) {
			--startline;
			size_t offset = 0;

			for(regmatch_t match; !regexec(&rx, line + offset, 1, &match, offset ? REG_NOTBOL : 0);) {
				regoff_t len = match.rm_eo - match.rm_so;
				if(urlcount >= urlsize) {
					size_t urlsizetmp = urlsize ? urlsize * 2 : 16;
					void * urltmp     = reallocarray(url, urlsizetmp, sizeof(*url));
					if(urltmp == NULL) {
						warn(gettext("couldn't allocate memory for additional URLs, only first %zu displayed"), urlsize);
						goto got_urls;
					} else {
						urlsize = urlsizetmp;
						url     = urltmp;
					}
				}
				url[urlcount].url = malloc(len + 1);
				memcpy(url[urlcount].url, line + match.rm_so + offset, len);
				url[urlcount].url[len] = 0;

				for(size_t i = 0; i < urlcount; ++i)
					if(!strcasecmp(url[i].url, url[urlcount].url)) {
						--urlcount;
						break;
					}

				if(current == (size_t)-1 && startline <= 0)
					current = urlcount;
				++urlcount;
				offset += match.rm_eo;
			}
		}
	got_urls:
		if(fp != stdin)
			fclose(fp);
	}
	regfree(&rx);
	free(line);

	if(getenv("URLVIEW_DEBUG") || !isatty(1)) {
		for(size_t i = 0; i < urlcount; ++i)
			puts(url[i].url);
		return 0;
	}
	if(!urlcount) {
		puts(gettext("No URLs found."));
		return 1;
	}

	if(current == (size_t)-1)
		current = urlcount - 1;

	if(reopen_tty)
		if(!freopen("/dev/tty", "r", stdin))
			err(1, "/dev/tty");
	initscr();

	cbreak();
	noecho();
	curs_set(1);
	keypad(stdscr, TRUE);
	if(mouse) {
// OpenBSD shim: it ships with NCURSES_MOUSE_VERSION being 1, which excludes BUTTON[45]_PRESSED: https://builds.sr.ht/~nabijaczleweli/job/1064539
#ifndef BUTTON5_PRESSED
#define BUTTON5_PRESSED 0
#endif
#ifndef BUTTON4_PRESSED
#define BUTTON4_PRESSED 0
#endif
		mousemask(BUTTON1_CLICKED | BUTTON5_PRESSED | BUTTON4_PRESSED, NULL);
		mouseinterval(0);
	}

//                                      title line
#define urlswin_logical_height (LINES - 1 - 1)
	//                                        status line
	WINDOW * urlswin = newpad(LINES, COLS);
	if(!urlswin)
		return endwin(), 1;

	enum redraw redraw   = FULL;
	char * search        = NULL;
	regex_t search_rx    = {.re_nsub = (size_t)-1};
	int page_of_current  = 0;
	size_t first_on_page = 0;
	int fudge            = 0;
	for(bool done = false; !done;) {
		switch(redraw) {
			case NO:
				break;
			case FULL: {
				wresize(urlswin, LINES, COLS);

				clear();
				standout();

				// Dickey recommends this! https://stackoverflow.com/a/70590135/2851815
				// Using urlswin since it's cleared below anyway.
				cc_t intr = CINTR;
				struct termios termios;
				if(!tcgetattr(0, &termios))
					intr = termios.c_cc[VINTR];
				char intr_s[4] = {};  // M-^?
				if(intr != _POSIX_VDISABLE) {
					char * intr_cur = intr_s;
					if(intr & 0x80)
						*intr_cur++ = 'M', *intr_cur++ = '-';
					intr &= 0x7F;
					if(isprint(intr))
						*intr_cur++ = intr;
					else if(intr == 0x7F)
						*intr_cur++ = '^', *intr_cur++ = '?';
					else
						*intr_cur++ = '^', *intr_cur++ = intr + '@';
					mvwprintw(urlswin, 0, 0, gettext("Press q or %.*s to quit!"), (int)sizeof(intr_s), intr_s);
				} else
					mvwaddstr(urlswin, 0, 0, gettext("Press q to quit!"));

#define BRAND "urlview-ng " VERSION ""
				mvwaddstr(stdscr, 0, COLS / 2 - (sizeof(BRAND) - 1) / 2, BRAND);
				if(intr != _POSIX_VDISABLE)
					mvwprintw(stdscr, 0, COLS - getcurx(urlswin), gettext("Press q or %.*s to quit!"), (int)sizeof(intr_s), intr_s);
				else
					mvwaddstr(stdscr, 0, COLS - getcurx(urlswin), gettext("Press q to quit!"));
				mvwprintw(stdscr, 0, 0, ngettext("%zu URL", "%zu URLs", urlcount), urlcount);
				standend();
				wnoutrefresh(stdscr);
			}
				__attribute__((fallthrough));
			case INDEX:
				wclear(urlswin);
				wmove(urlswin, 0, 0);
				for(size_t i = 0; i < urlcount; ++i) {
#define EXPAND()                                  \
	{                                               \
		int sizey, sizex;                             \
		getmaxyx(urlswin, sizey, sizex);              \
		if(wresize(urlswin, sizey * 2, sizex) == ERR) \
			break;                                      \
	}
#define EXPANDONSCROLL(...)         \
	if(__VA_ARGS__ != OK) {           \
		wmove(urlswin, starty, startx); \
		EXPAND();                       \
		goto again;                     \
	}
					int starty, startx;
					getyx(urlswin, starty, startx);

				again:
					EXPANDONSCROLL(waddstr(urlswin, "   "));
					getyx(urlswin, url[i].cursor_y, url[i].cursor_x);
					EXPANDONSCROLL(wprintw(urlswin, "%4zu", i + 1));
					EXPANDONSCROLL(waddch(urlswin, ' '));
					EXPANDONSCROLL(waddstr(urlswin, url[i].url));
					url[i].last_y = getcury(urlswin);

					if(i != urlcount - 1)
						if(waddch(urlswin, '\n') != OK) {
							EXPAND();
							waddch(urlswin, '\n');
						}
				}
				__attribute__((fallthrough));
			case MOTION:
				mvwaddstr(urlswin, url[oldcurrent].cursor_y, url[oldcurrent].cursor_x - 3, "  ");
				wstandout(urlswin);
				mvwaddstr(urlswin, url[current].cursor_y, url[current].cursor_x - 3, "->");
				wstandend(urlswin);

				page_of_current = url[current].cursor_y / urlswin_logical_height;
				first_on_page   = 0;
				while(url[first_on_page].cursor_y / urlswin_logical_height != page_of_current)
					++first_on_page;
				fudge = 0;  // try to get the last-URL-on-the-page to be included when it's selected
				if(url[current].last_y - url[first_on_page].cursor_y > urlswin_logical_height)
					// unclear why +1 needed here; this means we overscroll (sometimes)
					// compare the first screen on a rows 54; cols 172; teletype of
					//   ./urlview text.uv <(for i in A B C D E F H I J K L M; do tr Q $i < text.uv; done)
					// and
					//   ./urlview text.uv <(for i in A B C D E F H I J K L M; do tr Q $i < text.uv; done | fold -w 300)
					// the last link on the first screen appear to have a completely identical last_y, but the former is taller by a line?
					fudge = (url[current].last_y - url[first_on_page].cursor_y) - urlswin_logical_height + 1;
				pnoutrefresh(urlswin, /**/ url[first_on_page].cursor_y + fudge, 0, /**/ 1 /*title line*/, 0, urlswin_logical_height, COLS);
				doupdate();

				oldcurrent = current;
				break;
		}
		redraw = NO;

		int c = getch();
	rekey:
		switch(c) {
			case KEY_RESIZE:
				redraw = FULL;
				break;
			case KEY_MOUSE: {
				MEVENT ev;
				getmouse(&ev);

				if(ev.bstate & (BUTTON1_CLICKED | BUTTON1_DOUBLE_CLICKED | BUTTON1_TRIPLE_CLICKED)) {
					wmouse_trafo(urlswin, &ev.y, &ev.x, false);
					ev.y += url[first_on_page].cursor_y + fudge;

					for(size_t hit = first_on_page; hit != urlcount; ++hit)
						if(ev.y >= url[hit].cursor_y && ev.y <= url[hit].last_y) {
							current = hit;
							redraw  = MOTION;

							ungetch('\n');
							break;
						}
				} else if(ev.bstate & BUTTON5_PRESSED) {
					c = KEY_DOWN;
					goto rekey;
				} else if(ev.bstate & BUTTON4_PRESSED) {
					c = KEY_UP;
					goto rekey;
				}
			} break;
			case 'q':
			case 'x':
			case 'h':
				done = true;
				break;
			case KEY_DOWN:
			case 'j':
				if(current < urlcount - 1)
					++current;
				else {
					if(menu_wrapping)
						current = 0;
				}
				redraw = MOTION;
				break;
			case KEY_UP:
			case 'k':
				if(current)
					--current;
				else {
					if(menu_wrapping)
						current = urlcount - 1;
				}
				redraw = MOTION;
				break;
			case KEY_HOME:
			case '=':
				current = 0;
				redraw  = MOTION;
				break;
			case KEY_END:
			case '*':
			case 'G':
				current = urlcount - 1;
				redraw  = MOTION;
				break;
			case KEY_NPAGE:
			case '\006': {
				size_t first_on_next_page = first_on_page;
				while(first_on_next_page != urlcount - 1 && url[first_on_next_page].cursor_y / urlswin_logical_height != page_of_current + 1)
					++first_on_next_page;
				current = first_on_next_page;
				redraw  = MOTION;
			} break;
			case KEY_PPAGE:
			case '\002':
				if(first_on_page != 0)
					current = first_on_page - 1;
				else
					current = 0;
				redraw = MOTION;
				break;
			case '\n':
			case '\r':
			case KEY_ENTER:
			case ' ':
			case 'e':
				// edit prompt
				if(edit_url || c == 'e') {
					mvaddstr(LINES - 1, 0, gettext("URL: "));
					mmask_t oldmouse = 0;
					mousemask(0, &oldmouse);
					char * newurl = mutt_enter_string(url[current].url);
					if(newurl && *newurl) {
						if(strcmp(url[current].url, newurl))
							redraw = INDEX;
						free(url[current].url);
						url[current].url = newurl;
					}
					mousemask(oldmouse, NULL);
					move(LINES - 1, 0);
					clrtoeol();
					if(c == 'e' || !(newurl && *newurl))
						break;
				}
				endwin();


				char * execstatus    = NULL;
				size_t execstatuslen = 0;
#define EXECAPPENDONE(str, len)                              \
	{                                                          \
		char * new = realloc(execstatus, execstatuslen + (len)); \
		if(new) {                                                \
			execstatus = new;                                      \
			memcpy(execstatus + execstatuslen, str, (len));        \
			execstatuslen += (len);                                \
		}                                                        \
	}
#define EXECAPPEND_PCNT_S(fmt, s)              \
	{                                            \
		const char * pcnt = strstr(fmt, "%s");     \
		EXECAPPENDONE(fmt, pcnt - fmt);            \
		EXECAPPENDONE(s, strlen(s));               \
		EXECAPPENDONE(pcnt + 2, strlen(pcnt + 2)); \
	}

				bool okexit      = false;
				char * curcom    = command;
				size_t curcomlen = strcspn(curcom, ":");
				while(*curcom) {
					char * toexec = quotesub(curcom, curcomlen, url[current].url);
					EXECAPPEND_PCNT_S(gettext("Executing: %s... "), toexec);
					printf(gettext("Executing: %s... "), toexec);
					fflush(stdout);
					int childret = system(toexec);
					free(toexec);
					if(WIFEXITED(childret) && !WEXITSTATUS(childret)) {
						EXECAPPENDONE("i", 1);
						putchar('\n');
						okexit = true;
						break;
					}

					char childrets_buf[11 + 1], *childrets = childrets_buf;  // -2147483648
					if(WIFSIGNALED(childret))
						childrets = strsignal(WTERMSIG(childret));
					else
						sprintf(childrets_buf, "%d", childret);

					// child exit code or human-readable signal name, written after "Executing: %s... "
					EXECAPPEND_PCNT_S(gettext("%s! "), childrets);
					printf(gettext("%s!\n"), childrets);

					curcom += curcomlen;
					if(*curcom)
						++curcom;
					curcomlen = strcspn(curcom, ":");
				}
				if(execstatus) {
					size_t off = 0;
					for(; off != execstatuslen && mvaddnstr(LINES - 1, 0, execstatus + off, execstatuslen - off) == ERR; ++off)
						;
					if(okexit) {
						mvaddch(LINES - 1, getcurx(stdscr) - 1, ACS_DIAMOND);
						clrtoeol();
						addch('\b');
					}
					free(execstatus);
				}
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9': {
				// edit prompt
				mvaddstr(LINES - 1, 0, gettext("Jump to URL: "));
				char * num = mutt_enter_string((char[2]){c, '\0'});
				if(num && *num) {
					int i = atoi(num);
					free(num);
					if(i < 1 || (size_t)i > urlcount) {
						// error from the above
						mvaddstr(LINES - 1, 0, gettext("No such URL!"));
						clrtoeol();
					} else {
						current = i - 1;
						redraw  = MOTION;
					}
				}
			} break;
			case '\f':
			case '\007':
				clearok(stdscr, TRUE);
				redraw = FULL;
				break;
			case 'n':
				if(search_rx.re_nsub != (size_t)-1) {
					search_common(NULL, &search_rx, urlcount, url, &redraw, &current, search_forward);
					break;
				}
				__attribute__((fallthrough));
			case '/':
				// edit prompt
				mvaddstr(LINES - 1, 0, gettext("Search forwards for: "));
				char * newsearch = mutt_enter_string(search);
				if(newsearch && *newsearch) {
					free(search);
					search = newsearch;
					search_common(search, &search_rx, urlcount, url, &redraw, &current, search_forward);
				} else {
					move(LINES - 1, 0);
					clrtoeol();
				}
				break;
			case 'N':
				if(search_rx.re_nsub != (size_t)-1) {
					search_common(NULL, &search_rx, urlcount, url, &redraw, &current, search_backward);
					break;
				}
				__attribute__((fallthrough));
			case '?':
				// edit prompt
				mvaddstr(LINES - 1, 0, gettext("Search backwards for: "));
				newsearch = mutt_enter_string(search);
				if(newsearch && *newsearch) {
					free(search);
					search = newsearch;
					search_common(search, &search_rx, urlcount, url, &redraw, &current, search_backward);
				} else {
					move(LINES - 1, 0);
					clrtoeol();
				}
				break;
			default:
				break;
		}
	}

	endwin();
	return error;
}
